/*
Einfaches netzbetriebenes (stationres) Auto-Range-Voltmeter fr 
Gleich- und Wechselspannung mit:

- Atmel-AVR ATTiny24 als Controller mit ADC
- Dual-OPV als Messgleichrichter
- zwei Kleinrelais zur Umschaltung zwischen 4 Messbereichen
- Text-LCD mit 1 Zeile zu 10 Zeichen (Pollin 120593)
- Text-LCD mit 2 Zeilen zu 16 Zeichen zur Inbetriebnahme

Das Programm fr den Tiny24 wird vorerst nur als Hex-Datei verffentlicht.
Es steht bisher in zwei Varianten zur Verfgung:

VM24_LCD10x1.hex ist fr das finale Gert mit LCD 10x1 von Pollin.

VM24_LCD16x2.hex ist fr die Inbetriebnahme gedacht und arbeitet mit
einem LCD 16x2. Hier werden zustzlich zum eigentlichen Messwert noch
die Zwischenwerte an beiden ADC-Eingngen und der aktuelle Messbereich 
angezeigt. Dies dient der Inbetriebnahme des Analogteils.

Der ATTiny24 soll mit 1 MHz internem Takt (8 MHz und aktive CKDIV8-Fuse)
betrieben werden, BOD-Reset sollte per Fuse auf 2,3 V eingestellt werden
und das Aktivieren der EESAVE-Fuse schadet auch nicht.

-----------------------------------------------------------------------------

Um die Kalibration des Voltmeters auch ohne Offenlegung des Programm-
Quelltextes zu ermglichen, wurde die Kalibration in ein eigenes Projekt
ausgelagert, das nur die EEP-Datei fr das EEPROM des ATTiny24 erzeugt,
in der nur die 6 Kalibrierwerte liegen.

Jeder der 4 Messbereiche lsst sich separat in 0,4%-Stufen nach oben oder 
unten trimmen, um Toleranzen am Eingangs-Spannungsteiler kompensieren zu
knnen.

Durch Trimmen der DC-Anzeige und AC-Anzeige knnen Toleranzen der
AVR-internen Referenzspannung sowie Abweichungen des Messgleichrichters
kompensiert werden.

Sinnvoll sind Kalibrationswerte im Bereich von -32 bis +32, meist jedoch
bedeutend kleiner.

Zum Kalibrieren sind die Parameter anzupassen, dies Projekt zu assemblieren
(AVR-Studio4, AVR-Assembler2) und die Datei "VM24l_kalibration.eep" per ISP
in den EEPROM des ATTiny24 zu bertragen.

(c) 08/2009 by www.hanneslux.de (hannes.lux@gmx.de)
*/

.equ kali8=     0       ;Kalibrierwert fr Messbereich 8V 
.equ kali32=    0       ;Kalibrierwert fr Messbereich 32V 
.equ kali128=   0       ;Kalibrierwert fr Messbereich 128V
.equ kali512=   0       ;Kalibrierwert fr Messbereich 512V
.equ kalid=     -1      ;Kalibrierwert fr DC-Anzeige
.equ kalia=     22      ;Kalibrierwert fr AC-Anzeige


.eseg
.org 0
kal_mb:     .db kali8
            .db kali32
            .db kali128
            .db kali512
kal_dc:     .db kalid
kal_ac:     .db kalia

ctxt:
.db "  AVR-Mini-Voltmeter (C) 08/2009 by Hannes Lux www.hanneslux.de  "
